#include "RC4.h"

/*************************************************************/

unsigned int __declspec(naked) TimeIt(unsigned char Key[5])
{
	__asm {
		mov		ecx, [esp + 0x04]

/*
 * Define rdtsc intruction
 */

#define rdtsc __asm {						\
	__asm	_emit	0x0F					\
	__asm	_emit	0x31					\
}


		rdtsc
		push	edx
		push	eax
		push	ecx
		call	TrueRC4PartialExhaustion
		pop		ecx
		rdtsc
		pop		ecx
		sub		eax, ecx
		pop		ecx
		sbb		edx, ecx
		sub		eax, 18
		sbb		edx, 0		
		ret
	}
}		

/*************************************************************/

bool SlowRC4PartialExhaustion(unsigned char Key[5])
{
	unsigned int I, J, A, B, S[256];
	Key[4] = 0x00;	

	do {
		for (I = 0; I < 256; I++)
			S[I] = I;

		for (I = 0, J = 0; I < 256; I++) {
			A = S[I];
			J = (J + A + Key[I % 5]) % 256;
			B = S[J];
			S[J] = A;
			S[I] = B;
		}

	} while (Key[4]++ != 0xFF);

	return false;
}

/*************************************************************/

bool FastRC4PartialExhaustion(unsigned char Key[5])
{
	unsigned int I, K;
	unsigned int Key1[5], J1, A1, B1, S1[256];
	unsigned int Key2[5], J2, A2, B2, S2[256];

	Key1[0] = Key2[0] = Key[0];
	Key1[1] = Key2[1] = Key[1];
	Key1[2] = Key2[2] = Key[2];
	Key1[3] = Key2[3] = Key[3];

	for (Key1[4] = 0, Key2[4] = 1; Key1[4] < 256; Key1[4] += 2, Key2[4] += 2) {

		for (I = 0; I < 256; I++) {
			S1[I] = I;
			S2[I] = I;
		}

		for (I = 0, J1 = J2 = 0, K = 0; I < 256; I++) {
			A1 = S1[I];
			A2 = S2[I];

			J1 += Key1[K];
			J2 += Key2[K];

			J1 += A1;
			J2 += A2;

			J1 &= 0xFF;
			J2 &= 0xFF;

			B1 = S1[J1];
			B2 = S2[J2];

			S1[J1] = A1;
			S2[J2] = A2;

			S1[I] = B1;
			S2[I] = B2;

			if (++K > 4)
				K = 0;
		}

		// Generate keystream.
	}

	return false;
}

/*************************************************************/

bool __declspec(naked) TrueRC4PartialExhaustion(unsigned char Key[5])
{
	__asm {

/*
 * Save a copy of the first two bytes of Key, and set the
 * three other bytes to zero. Should take 5 cpu cycles
 */

		mov		ecx, [esp + 0x04]			// Key
		push	ebp

		push	ebx
		push	edi

		mov		edx, [ecx]					// First two bytes in DX
		push	esi
		
//		and		edx, 0x0000FFFF				// Zero bytes three & four
		nop
		push	0x0000FEFF					// Two copies of byte 5

		push	edx							// Save bytes 0 - 3 of key
		mov		ebp, esp

		xor		eax, eax
		xor		ebx, ebx

/*
 * Thus at that point, we have the following memory usage:
 *    [EBP + 0]      First byte of cryptovariable
 *    [EBP + 1]      Second byte of cryptovariable
 *    [EBP + 2]      Third byte of cryptovariable (Set to 0x00)
 *    [EBP + 3]      Fourth byte of cryptovariable (Set to 0x00)
 *    [EBP + 4]      Fifth byte of cryptovariable, copy A (Set to 0xFF)
 *    [EBP + 5]      Fourth byte of cryptovariable, copy B (Set to 0xFE)
 *    [EBP + 6]      Junk (0x0000)
 */

/*
 * Set an identity permutation. Should take 128 cpu cycles.
 */

		push	0x0000FFFF
		push	0x0000FEFE
		push	0x0000FDFD
		push	0x0000FCFC
		push	0x0000FBFB
		push	0x0000FAFA
		push	0x0000F9F9
		push	0x0000F8F8
		push	0x0000F7F7
		push	0x0000F6F6
		push	0x0000F5F5
		push	0x0000F4F4
		push	0x0000F3F3
		push	0x0000F2F2
		push	0x0000F1F1
		push	0x0000F0F0
		push	0x0000EFEF
		push	0x0000EEEE
		push	0x0000EDED
		push	0x0000ECEC
		push	0x0000EBEB
		push	0x0000EAEA
		push	0x0000E9E9
		push	0x0000E8E8
		push	0x0000E7E7
		push	0x0000E6E6
		push	0x0000E5E5
		push	0x0000E4E4
		push	0x0000E3E3
		push	0x0000E2E2
		push	0x0000E1E1
		push	0x0000E0E0
		push	0x0000DFDF
		push	0x0000DEDE
		push	0x0000DDDD
		push	0x0000DCDC
		push	0x0000DBDB
		push	0x0000DADA
		push	0x0000D9D9
		push	0x0000D8D8
		push	0x0000D7D7
		push	0x0000D6D6
		push	0x0000D5D5
		push	0x0000D4D4
		push	0x0000D3D3
		push	0x0000D2D2
		push	0x0000D1D1
		push	0x0000D0D0
		push	0x0000CFCF
		push	0x0000CECE
		push	0x0000CDCD
		push	0x0000CCCC
		push	0x0000CBCB
		push	0x0000CACA
		push	0x0000C9C9
		push	0x0000C8C8
		push	0x0000C7C7
		push	0x0000C6C6
		push	0x0000C5C5
		push	0x0000C4C4
		push	0x0000C3C3
		push	0x0000C2C2
		push	0x0000C1C1
		push	0x0000C0C0
		push	0x0000BFBF
		push	0x0000BEBE
		push	0x0000BDBD
		push	0x0000BCBC
		push	0x0000BBBB
		push	0x0000BABA
		push	0x0000B9B9
		push	0x0000B8B8
		push	0x0000B7B7
		push	0x0000B6B6
		push	0x0000B5B5
		push	0x0000B4B4
		push	0x0000B3B3
		push	0x0000B2B2
		push	0x0000B1B1
		push	0x0000B0B0
		push	0x0000AFAF
		push	0x0000AEAE
		push	0x0000ADAD
		push	0x0000ACAC
		push	0x0000ABAB
		push	0x0000AAAA
		push	0x0000A9A9
		push	0x0000A8A8
		push	0x0000A7A7
		push	0x0000A6A6
		push	0x0000A5A5
		push	0x0000A4A4
		push	0x0000A3A3
		push	0x0000A2A2
		push	0x0000A1A1
		push	0x0000A0A0
		push	0x00009F9F
		push	0x00009E9E
		push	0x00009D9D
		push	0x00009C9C
		push	0x00009B9B
		push	0x00009A9A
		push	0x00009999
		push	0x00009898
		push	0x00009797
		push	0x00009696
		push	0x00009595
		push	0x00009494
		push	0x00009393
		push	0x00009292
		push	0x00009191
		push	0x00009090
		push	0x00008F8F
		push	0x00008E8E
		push	0x00008D8D
		push	0x00008C8C
		push	0x00008B8B
		push	0x00008A8A
		push	0x00008989
		push	0x00008888
		push	0x00008787
		push	0x00008686
		push	0x00008585
		push	0x00008484
		push	0x00008383
		push	0x00008282
		push	0x00008181
		push	0x00008080
		push	0x00007F7F
		push	0x00007E7E
		push	0x00007D7D
		push	0x00007C7C
		push	0x00007B7B
		push	0x00007A7A
		push	0x00007979
		push	0x00007878
		push	0x00007777
		push	0x00007676
		push	0x00007575
		push	0x00007474
		push	0x00007373
		push	0x00007272
		push	0x00007171
		push	0x00007070
		push	0x00006F6F
		push	0x00006E6E
		push	0x00006D6D
		push	0x00006C6C
		push	0x00006B6B
		push	0x00006A6A
		push	0x00006969
		push	0x00006868
		push	0x00006767
		push	0x00006666
		push	0x00006565
		push	0x00006464
		push	0x00006363
		push	0x00006262
		push	0x00006161
		push	0x00006060
		push	0x00005F5F
		push	0x00005E5E
		push	0x00005D5D
		push	0x00005C5C
		push	0x00005B5B
		push	0x00005A5A
		push	0x00005959
		push	0x00005858
		push	0x00005757
		push	0x00005656
		push	0x00005555
		push	0x00005454
		push	0x00005353
		push	0x00005252
		push	0x00005151
		push	0x00005050
		push	0x00004F4F
		push	0x00004E4E
		push	0x00004D4D
		push	0x00004C4C
		push	0x00004B4B
		push	0x00004A4A
		push	0x00004949
		push	0x00004848
		push	0x00004747
		push	0x00004646
		push	0x00004545
		push	0x00004444
		push	0x00004343
		push	0x00004242
		push	0x00004141
		push	0x00004040
		push	0x00003F3F
		push	0x00003E3E
		push	0x00003D3D
		push	0x00003C3C
		push	0x00003B3B
		push	0x00003A3A
		push	0x00003939
		push	0x00003838
		push	0x00003737
		push	0x00003636
		push	0x00003535
		push	0x00003434
		push	0x00003333
		push	0x00003232
		push	0x00003131
		push	0x00003030
		push	0x00002F2F
		push	0x00002E2E
		push	0x00002D2D
		push	0x00002C2C
		push	0x00002B2B
		push	0x00002A2A
		push	0x00002929
		push	0x00002828
		push	0x00002727
		push	0x00002626
		push	0x00002525
		push	0x00002424
		push	0x00002323
		push	0x00002222
		push	0x00002121
		push	0x00002020
		push	0x00001F1F
		push	0x00001E1E
		push	0x00001D1D
		push	0x00001C1C
		push	0x00001B1B
		push	0x00001A1A
		push	0x00001919
		push	0x00001818
		push	0x00001717
		push	0x00001616
		push	0x00001515
		push	0x00001414
		push	0x00001313
		push	0x00001212
		push	0x00001111
		push	0x00001010
		push	0x00000F0F
		push	0x00000E0E
		push	0x00000D0D
		push	0x00000C0C
		push	0x00000B0B
		push	0x00000A0A
		push	0x00000909
		push	0x00000808
		push	0x00000707
		push	0x00000606
		push	0x00000505
		push	0x00000404
		push	0x00000303
		push	0x00000202
		push	0x00000101
		push	0x00000000

		jmp		OddRunup

/*
 * After this first initialization, for (0 <= i < 256), we have
 *    [EDI + 4 * i]       Ith element of permutation S, Copy A
 *    [EDI + 4 * i + 1]   Ith element of permutation S, Copy B
 *    [EDI + 4 * i + 2]   Junk (0x0000)
 */

/*
 * Register usage:
 *    AL       Counter J, Copy A
 *    BL       Counter J, Copy B
 *    CX       S[I], Copy A and Copy B
 *    EDX      Cryptovariable bytes (usually in DL, but in DX for both copies of fith byte)
 *    EDI      Pointer to permutations array
 *    ESI      &S[I]
 *    EBP      Pointer to cryptovariable
 *    ESP      Pointer to next permutation arrays
 */

/*
 * Runup. First iteration.
 * Cpu cycles: 6
 */

#define iunup(ToTop) __asm {				\
	__asm	mov		al, dl					\
	__asm	mov		bl, dl					\
											\
	__asm	mov		[esp], dl				\
	__asm	mov		edi, esp				\
											\
	__asm	mov		[esp + 1], dl			\
	__asm	mov		esi, esp				\
											\
	__asm	mov		[edi + 4 * eax], ah		\
	__asm	add		esp, ToTop * 2048		\
											\
	__asm	mov		[edi + 4 * ebx + 1], bh	\
	__asm	add		al, dh					\
											\
	__asm	push	0x0000FFFF				\
	__asm	add		bl, dh					\
}

/*
 * Runup iteration, 1 <= I <= 255 but (I mod 5) != 3
 * Cpu cycles: 8
 */

 #define runup(I) __asm {					\
	__asm	mov		ecx, [esi + 4]			\
	__asm	push	0xFFFF - ((I) * 0x0101)						\
											\
	__asm	add		al, cl					\
	__asm	add		bl, ch					\
											\
	__asm	mov		dl, [ebp + (I + 1) % 5]	\
	__asm	push	ecx						\
											\
	__asm	mov		cl, [edi + 4 * eax]		\
	__asm	add		esi, 4					\
											\
	__asm	mov		ch, [edi + 4 * ebx + 1]	\
	__asm	nop								\
											\
	__asm	mov		[esi], ecx				\
	__asm	pop		ecx						\
											\
	__asm	mov		[edi + 4 * eax], cl		\
	__asm	add		al, dl					\
											\
	__asm	mov		[edi + 4 * ebx + 1], ch	\
	__asm	add		bl, dl					\
}

/*
 * Runup iteration, 1 <= I <= 255 and (I mod 5) == 3
 * Cpu cycles: 8
 */

 #define xunup(I) __asm {					\
	__asm	mov		ecx, [esi + 4]			\
	__asm	push	(0x10100 - (I * 0x101))	\
											\
	__asm	add		al, cl					\
	__asm	add		bl, ch					\
											\
	__asm	mov		edx, [ebp + 4]			\
	__asm	push	ecx						\
											\
	__asm	mov		cl, [edi + 4 * eax]		\
	__asm	add		esi, 4					\
											\
	__asm	mov		ch, [edi + 4 * ebx + 1]	\
	__asm	nop								\
											\
	__asm	mov		[esi], ecx				\
	__asm	pop		ecx						\
											\
	__asm	mov		[edi + 4 * eax], cl		\
	__asm	add		al, dl					\
											\
	__asm	mov		[edi + 4 * ebx + 1], ch	\
	__asm	add		bl, dh					\
}

/*
 * Complete runup
 */

OddRunup:
		iunup(0);
		jmp		DoRunup
EvenRunup:
		iunup(1);
		jmp		DoRunup

/*
 * We may need to play with ESP...
 */

		
DoRunup:
		runup(1)
		runup(2)
		xunup(3)
		runup(4)
		runup(5)
		runup(6)
		runup(7)
		xunup(8)
		runup(9)
		runup(10)
		runup(11)
		runup(12)
		xunup(13)
		runup(14)
		runup(15)
		runup(16)
		runup(17)
		xunup(18)
		runup(19)
		runup(20)
		runup(21)
		runup(22)
		xunup(23)
		runup(24)
		runup(25)
		runup(26)
		runup(27)
		xunup(28)
		runup(29)
		runup(30)
		runup(31)
		runup(32)
		xunup(33)
		runup(34)
		runup(35)
		runup(36)
		runup(37)
		xunup(38)
		runup(39)
		runup(40)
		runup(41)
		runup(42)
		xunup(43)
		runup(44)
		runup(45)
		runup(46)
		runup(47)
		xunup(48)
		runup(49)
		runup(50)
		runup(51)
		runup(52)
		xunup(53)
		runup(54)
		runup(55)
		runup(56)
		runup(57)
		xunup(58)
		runup(59)
		runup(60)
		runup(61)
		runup(62)
		xunup(63)
		runup(64)
		runup(65)
		runup(66)
		runup(67)
		xunup(68)
		runup(69)
		runup(70)
		runup(71)
		runup(72)
		xunup(73)
		runup(74)
		runup(75)
		runup(76)
		runup(77)
		xunup(78)
		runup(79)
		runup(80)
		runup(81)
		runup(82)
		xunup(83)
		runup(84)
		runup(85)
		runup(86)
		runup(87)
		xunup(88)
		runup(89)
		runup(90)
		runup(91)
		runup(92)
		xunup(93)
		runup(94)
		runup(95)
		runup(96)
		runup(97)
		xunup(98)
		runup(99)
		runup(100)
		runup(101)
		runup(102)
		xunup(103)
		runup(104)
		runup(105)
		runup(106)
		runup(107)
		xunup(108)
		runup(109)
		runup(110)
		runup(111)
		runup(112)
		xunup(113)
		runup(114)
		runup(115)
		runup(116)
		runup(117)
		xunup(118)
		runup(119)
		runup(120)
		runup(121)
		runup(122)
		xunup(123)
		runup(124)
		runup(125)
		runup(126)
		runup(127)
		xunup(128)
		runup(129)
		runup(130)
		runup(131)
		runup(132)
		xunup(133)
		runup(134)
		runup(135)
		runup(136)
		runup(137)
		xunup(138)
		runup(139)
		runup(140)
		runup(141)
		runup(142)
		xunup(143)
		runup(144)
		runup(145)
		runup(146)
		runup(147)
		xunup(148)
		runup(149)
		runup(150)
		runup(151)
		runup(152)
		xunup(153)
		runup(154)
		runup(155)
		runup(156)
		runup(157)
		xunup(158)
		runup(159)
		runup(160)
		runup(161)
		runup(162)
		xunup(163)
		runup(164)
		runup(165)
		runup(166)
		runup(167)
		xunup(168)
		runup(169)
		runup(170)
		runup(171)
		runup(172)
		xunup(173)
		runup(174)
		runup(175)
		runup(176)
		runup(177)
		xunup(178)
		runup(179)
		runup(180)
		runup(181)
		runup(182)
		xunup(183)
		runup(184)
		runup(185)
		runup(186)
		runup(187)
		xunup(188)
		runup(189)
		runup(190)
		runup(191)
		runup(192)
		xunup(193)
		runup(194)
		runup(195)
		runup(196)
		runup(197)
		xunup(198)
		runup(199)
		runup(200)
		runup(201)
		runup(202)
		xunup(203)
		runup(204)
		runup(205)
		runup(206)
		runup(207)
		xunup(208)
		runup(209)
		runup(210)
		runup(211)
		runup(212)
		xunup(213)
		runup(214)
		runup(215)
		runup(216)
		runup(217)
		xunup(218)
		runup(219)
		runup(220)
		runup(221)
		runup(222)
		xunup(223)
		runup(224)
		runup(225)
		runup(226)
		runup(227)
		xunup(228)
		runup(229)
		runup(230)
		runup(231)
		runup(232)
		xunup(233)
		runup(234)
		runup(235)
		runup(236)
		runup(237)
		xunup(238)
		runup(239)
		runup(240)
		runup(241)
		runup(242)
		xunup(243)
		runup(244)
		runup(245)
		runup(246)
		runup(247)
		xunup(248)
		runup(249)
		runup(250)
		runup(251)
		runup(252)
		xunup(253)
		runup(254)
		runup(255)

/*
 * Ready to generate a few bytes of keystream...
 */

/*
 * Now, prepare for next iteration
 */

		mov		ecx, ebp
		mov		edx, [ebp + 04]
		
		sub		ecx, esp
		sub		edx, 0x0202

		mov		[ebp + 04], edx
		mov		edx, [ebp]

		cmp		ecx, 1024
		jne		EvenRunup

		mov		ecx, [ebp + 04]
		nop

		or		ecx, ecx
		jns		OddRunup

/*
 * Exhaustion finished
 */		

		add		esp, 1024
		add		esp, 8
		xor		eax, eax
		
		pop		esi
		pop		edi
		pop		ebx
		pop		ebp
		ret
	}
}

/*************************************************************/
/*************************************************************/
/*************************************************************/

